import re

import requests

VHOST_DB_PATH = '/home/s/st/staff/vhost/vhost-app.conf'
VHOST_DB_URL = 'https://www.ocf.berkeley.edu/~staff/vhost-app.conf'


def get_app_vhost_db():
    """Returns lines from the application vhost database. Loaded from the
    filesystem (if available), or from the web if not."""
    try:
        with open(VHOST_DB_PATH) as f:
            return f.read().splitlines()
    except IOError:
        # fallback to database loaded from web
        return requests.get(VHOST_DB_URL, timeout=20).text.split('\n')


def get_app_vhosts():
    """Returns a list of application virtual hosts in convenient format.

    >>> get_app_vhosts()
    {
    ...
        'ml.berkeley.edu': {
            'username': 'mlab',
            'socket': 'mlab'
            'aliases': [],
            'flags': []
        }
    ...
    }
    """
    def fully_qualify(host):
        """Fully qualifies a hostname (by appending .berkeley.edu) if it's not
        already fully-qualified."""
        return host if '.' in host else host + '.berkeley.edu'

    vhosts = {}

    for line in get_app_vhost_db():
        if not line or line.startswith('#'):
            continue

        fields = line.split(' ')

        if len(fields) < 5:
            flags = []
        else:
            flags = re.match(r'\[(.*)\]$', fields[4]).group(1).split(',')

        username, host, socket, aliases = fields[:4]

        vhosts[fully_qualify(username if host == '-' else host)] = {
            'username': username,
            'socket': socket if socket != '-' else username,
            'aliases': aliases.split(',') if aliases != '-' else [],
            'flags': flags,
        }

    return vhosts



import functools
from collections import namedtuple
from datetime import datetime

from ocflib.account.search import user_exists
from ocflib.account.search import user_is_group
from ocflib.account.utils import is_in_group
from ocflib.infra import mysql

WEEKDAY_QUOTA = 10
WEEKEND_QUOTA = 20
SEMESTERLY_QUOTA = 100

# Per BoD decision of 2017-04-24, the rules for making changes to the printing
# quota, (non-normatively) summarized, are:
#   - BoD alone normally holds sole authority to change the printing quota.
#   - During breaks, Cabinet can make temporary changes to this quota.
#   - At the next BoD meeting, BoD can ratify those temporary changes to make
#     them permanent. If it doesn't, then those changes automatically expire.

HAPPY_HOUR_QUOTA = 20
HAPPY_HOUR_START = datetime(2019, 5, 6)
HAPPY_HOUR_END = datetime(2019, 5, 17)


get_connection = functools.partial(mysql.get_connection,
                                   user='anonymous',
                                   password=None,
                                   db='ocfprinting')

UserQuota = namedtuple('UserQuota', (
    'user',
    'daily',
    'semesterly',
))


Job = namedtuple('Job', (
    'user',
    'time',
    'pages',
    'queue',
    'printer',
    'doc_name',
    'filesize',
))

Refund = namedtuple('Refund', (
    'user',
    'time',
    'pages',
    'staffer',
    'reason',
))


def daily_quota(day=None):
    """Return the daily quota for a given day.

    :param day: date object (defaults to today)
    """
    if day is None:
        day = datetime.today()

    if HAPPY_HOUR_START <= day and day < HAPPY_HOUR_END:
        return HAPPY_HOUR_QUOTA
    elif day.weekday() in {5, 6}:
        return WEEKEND_QUOTA
    else:
        return WEEKDAY_QUOTA


def get_quota(c, user):
    """Return a UserQuota representing the user's quota."""
    if is_in_group(user, 'opstaff'):
        return UserQuota(user, 500, 500)

    if not user_exists(user) or user_is_group(user):
        return UserQuota(user, 0, 0)

    c.execute(
        'SELECT `today`, `semester` FROM `printed` WHERE `user` = %s',
        (user,)
    )

    row = c.fetchone()
    if not row:
        row = {'today': 0, 'semester': 0}
    semesterly = max(0, SEMESTERLY_QUOTA - int(row['semester']))
    return UserQuota(
        user=user,
        daily=max(0, min(semesterly, daily_quota() - int(row['today']))),
        semesterly=semesterly,
    )


def _namedtuple_to_query(query, nt):
    """Return a filled-out query and arguments.

    The return value can be exploded and passed directly into execute.

    >>> query = 'INSERT INTO jobs ({}) VALUES ({});'
    >>> namedtuple_to_query(query, job)
    ('INSERT INTO jobs (`user`, `pages`) VALUES (%s, %s)', ('ckuehl', 42))
    """
    return (
        query.format(
            ', '.join('`{}`'.format(column) for column in nt._fields),
            ', '.join('%s' for _ in nt._fields),
        ),
        tuple(getattr(nt, column) for column in nt._fields),
    )


def add_job(c, job):
    """Add a new job to the database."""
    c.execute(*_namedtuple_to_query('INSERT INTO jobs ({}) VALUES ({})', job))


def add_refund(c, refund):
    """Add a new refund to the database."""
    c.execute(*_namedtuple_to_query('INSERT INTO refunds ({}) VALUES ({})', refund))



import pymysql


def get_connection(user,
                   password,
                   db,
                   host='mysql.ocf.berkeley.edu',
                   cursorclass=pymysql.cursors.DictCursor,
                   charset='utf8mb4',
                   autocommit=True,
                   **kwargs):
    """Returns a context-manager aware connection to MariaDB, with sensible defaults.

    While this function can be called directly, there are partial function
    in some ocflib modules that may be better suited for particular tasks."""

    return pymysql.connect(
        user=user,
        password=password,
        db=db,
        host=host,
        cursorclass=cursorclass,
        charset=charset,
        autocommit=autocommit,
        **kwargs
    )



"""Print Kanboard task information."""
import json
from collections import namedtuple

import requests


KANBOARD_ROOT = 'https://kanboard.ocf.berkeley.edu'


def request(usr, api_key, method, params):
    """Sends a request to the Kanboard API in JSON-RPC 2.0 format."""
    # The purpose of id is to be able to reorder responses to asynchronous,
    # batched requests. Any valid integer is OK here, since we're just making
    # one simple request.
    payload = json.dumps({'jsonrpc': '2.0', 'method': method, 'id': 1, 'params': params})
    return requests.post(
        '{}/jsonrpc.php'.format(KANBOARD_ROOT),
        data=payload,
        auth=(usr, api_key),
        timeout=10,
    )


class KanboardError(ValueError):
    pass


class KanboardTask(namedtuple('KanboardTask', ('number', 'title', 'creator', 'project'))):
    """A namedtuple representing a Kanboard task."""

    def __str__(self):
        return (
            'k#{self.number}: "{self.title}" | '
            '{self.project}, started by {self.creator} | '
            'https://ocf.io/k/{self.number}'
        ).format(self=self)

    @classmethod
    def from_number(cls, usr, api_key, num):
        """Gets information about a Kanboard task based on its number.

        Example usage:

            KanboardTask.from_number('jsonrpc',
                '19ffd9709d03ce50675c3a43d1c49c1ac207f4bc45f06c5b2701fbdf8929', 1)

        :param usr: either the Kanboard username corresponding to the api key for a user
            api key or 'jsonrpc' for the application api key
        :param api_key: a user api key (which can be found under My profile -> Actions -> API) or the
            application api key (which only admins can see)
        """

        task_resp = request(usr, api_key, 'getTask', {'task_id': num})
        if task_resp.status_code != 200:
            raise KanboardError(
                'Task request gave {}'.format(task_resp.status_code)
            )

        task = task_resp.json()['result']

        users_resp = request(usr, api_key, 'getProjectUsers', {'project_id': task['project_id']})
        if users_resp.status_code != 200:
            raise KanboardError(
                'Project request gave {}'.format(users_resp.status_code)
            )

        users = users_resp.json()['result']

        proj_resp = request(usr, api_key, 'getProjectById', {'project_id': task['project_id']})
        if proj_resp.status_code != 200:
            raise KanboardError(
                'Project request gave {}'.format(proj_resp.status_code)
            )

        proj = proj_resp.json()['result']

        return cls(
            number=task['id'],
            title=task['title'],
            creator=users[task['creator_id']],
            project=proj['name'],
        )
